from . import _version
from .task_runners import DaskTaskRunner  # noqa
from .client import PrefectDaskClient  # noqa
from .utils import get_dask_client, get_async_dask_client  # noqa

__version__ = _version.__version__
